/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>

using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::gui;
using namespace mrpt::slam;
using namespace mrpt::random;
using namespace std;


/** A versatile template to build vectors on-the-fly in a style close to MATLAB's  v=[a b c d ...]
  *  The first argument of the template is the vector length, and the second the type of the numbers.
  *  Some examples:
  *
  *  \code
  *    vector_double  = _v<4,double>(1.0,3.0,4.0,5.0);
  *    vector_float   = _v<2,float>(-8.12, 3e4);
  *  \endcode
  */
template <size_t N, typename T>
std::vector<T> _v(const T val1, ...)
{
	MRPT_COMPILE_TIME_ASSERT( N>0 )
	std::vector<T>	ret;
	ret.reserve(N);

	ret.push_back(val1);

	va_list args;
	va_start(args,val1);
	for (size_t i=0;i<N-1;i++)
		ret.push_back( va_arg(args,T) );

	va_end(args);
	return ret;
}

void simulateObsPredicts( 
	const vector_double &GTx, 
	const vector_double &GTy, 
	map<size_t,CPointPDFGaussian> &obs, 
	map<size_t,CPointPDFGaussian> &pred)
{
	const double max_R = 10;
	const double max_angle = DEG2RAD(90);
	const double SIGMA_EST = 0.1;
	
	const double SIGMA_R   = 0.05;
	const double SIGMA_ANG = DEG2RAD(3);

	const CPose3D robotPose( 1,0,DEG2RAD(0) );
	const CPose3D robotPose_pred( 1.05,-0.04,DEG2RAD(-4) );

	pred.clear();
	obs.clear();

	for (size_t i=0;i<GTx.size();i++)
	{
		CPoint3D  real_p( GTx[i], GTy[i] );
		CPoint3D  est_p = real_p;

		est_p.x_incr( randomGenerator.drawGaussian1D(0,SIGMA_EST) );
		est_p.y_incr( randomGenerator.drawGaussian1D(0,SIGMA_EST) );

		// Only add to obs / pred if it falls in range:
		double r,y,p;
		robotPose.sphericalCoordinates( est_p, r,y,p);
		if (r<max_R && fabs(y)<max_angle)
		{
			pred[i].mean = est_p; // - robotPose;
			pred[i].cov.unit();
			pred[i].cov*= square(SIGMA_EST);
		}
		
		robotPose.sphericalCoordinates( real_p, r,y,p);
		if (r<max_R && fabs(y)<max_angle)
		{
			obs[i].mean  = robotPose_pred + (real_p - robotPose);

			double r = obs[i].mean.norm();
			double a = atan2( obs[i].mean.y(), obs[i].mean.x() );

			CMatrixDouble22 H;
			H(0,0) =  cos(a);  H(0,1) = -r*sin(a);
			H(1,0) =  sin(a);  H(1,1) =  r*cos(a);
			
			CMatrixDouble22 S;
			S(0,0) = square( SIGMA_R );
			S(1,1) = square( SIGMA_ANG );

			r+=randomGenerator.drawGaussian1D(0,SIGMA_R);
			a+=randomGenerator.drawGaussian1D(0,SIGMA_ANG);

			obs[i].mean.x( cos(a)*r );
			obs[i].mean.y( sin(a)*r );

			CMatrixDouble22 cov22(false,false);
			H.multiply_HCHt(S, cov22);
			obs[i].cov = CMatrixDouble33(cov22);
		}
	}
}

/* ------------------------------------------------------------------------
					Test: data-association-demo
   ------------------------------------------------------------------------ */
void Test_dataAssociation( const size_t  nLandmarks )
{
	randomGenerator.randomize( 1234 );

	// Generate the ground truth:
	//---------------------------------------------
	vector_double GT_x(nLandmarks),GT_y(nLandmarks);

	randomGenerator.drawUniformVector( GT_x, 1.0,5.0 );
	randomGenerator.drawUniformVector( GT_y, -6.0,6.0 );


	// Simulate observations & predictions:
	//---------------------------------------------
	map<size_t,CPointPDFGaussian>  observations;
	map<size_t,CPointPDFGaussian>  predictions;

	simulateObsPredicts(GT_x,GT_y,observations, predictions);


	// Do data-association:
	//---------------------------------------------
	TDataAssociationResults		results;

	//TDataAssociationMethod method = assocNN;
	TDataAssociationMethod method = assocJCBB;

	CTicTac  tictac;
	tictac.Tic();

	const size_t N = 1; // 1000

	for (size_t i=0;i<N;i++)
		mrpt::slam::data_association(predictions, observations, results, method );

	cout << "Time for data association: " << 1e3*tictac.Tac()/N << " ms" << endl;


	// Draw all:
	// --------------------------------------------------------
	CDisplayWindowPlots	win2("Predictions (red) & observations (blue)");

	for (map<size_t,CPointPDFGaussian>::const_iterator itO=observations.begin();itO!=observations.end();++itO)
	{
		CMatrixDouble COV22 = CMatrixDouble(itO->second.cov);
		COV22.resize(2,2);

		win2.plotEllipse( itO->second.mean.x(), itO->second.mean.y(), COV22, 3, "b", format("obs_%u",(unsigned)itO->first));
		win2.plot( _v<2,double>(itO->second.mean.x(), GT_x[itO->first]), _v<2,double>(itO->second.mean.y(), GT_y[itO->first]), "k", format("gt_assoc_%u",(unsigned)itO->first));
	}

	for (map<size_t,CPointPDFGaussian>::const_iterator itP=predictions.begin();itP!=predictions.end();++itP)
	{
		CMatrixDouble COV22 = CMatrixDouble(itP->second.cov);
		COV22.resize(2,2);
		win2.plotEllipse( itP->second.mean.x(), itP->second.mean.y(), COV22, 3, "r", format("pred_%u",(unsigned)itP->first));
	}
	win2.plot(GT_x,GT_y,".k5","landmarks");

	win2.axis_fit(true);
	win2.setPos(10,10);
	win2.resize(400,400);


	// Draw individual compatibility
	// --------------------------------------------------------
	CDisplayWindowPlots	win1("Individual compatibility");
	for (map<size_t,CPointPDFGaussian>::const_iterator itO=observations.begin();itO!=observations.end();++itO)
	{
		CMatrixDouble COV22 = CMatrixDouble(itO->second.cov);
		COV22.resize(2,2);
		win1.plotEllipse( itO->second.mean.x(), itO->second.mean.y(), COV22, 3, "b", format("obs_%u",(unsigned)itO->first));
	}
	for (map<size_t,CPointPDFGaussian>::const_iterator itP=predictions.begin();itP!=predictions.end();++itP)
	{
		CMatrixDouble COV22 = CMatrixDouble(itP->second.cov);
		COV22.resize(2,2);
		win1.plotEllipse( itP->second.mean.x(), itP->second.mean.y(), COV22, 3, "r", format("pred_%u",(unsigned)itP->first));
	}
	for (size_t p=0;p<results.indiv_compatibility.getRowCount();p++)
		for (size_t o=0;o<results.indiv_compatibility.getColCount();o++)
			if (results.indiv_compatibility(p,o))
			{
				win1.plot( 
					_v<2,double>( observations[results.obsidx2id[o]].mean.x(), predictions[results.predidx2id[p]].mean.x() ), 
					_v<2,double>( observations[results.obsidx2id[o]].mean.y(), predictions[results.predidx2id[p]].mean.y() ), 
					"k", format("ic_%u_%u",(unsigned)p,(unsigned)o));
			}

	win1.axis_fit(true);
	win1.setPos(430,10);
	win1.resize(400,400);

	// Draw final associations
	// --------------------------------------------------------
	CDisplayWindowPlots	win3("Final data association");
	for (map<size_t,CPointPDFGaussian>::const_iterator itO=observations.begin();itO!=observations.end();++itO)
	{
		CMatrixDouble COV22 = CMatrixDouble(itO->second.cov);
		COV22.resize(2,2);
		win3.plotEllipse( itO->second.mean.x(), itO->second.mean.y(), COV22, 3, "b", format("obs_%u",(unsigned)itO->first));
	}
	for (map<size_t,CPointPDFGaussian>::const_iterator itP=predictions.begin();itP!=predictions.end();++itP)
	{
		CMatrixDouble COV22 = CMatrixDouble(itP->second.cov);
		COV22.resize(2,2);
		win3.plotEllipse( itP->second.mean.x(), itP->second.mean.y(), COV22, 3, "r", format("pred_%u",(unsigned)itP->first));
	}
	cout << results.associations.size() <<  " associations: " << endl;
	for (map<size_t,size_t>::const_iterator assoc = results.associations.begin();assoc!=results.associations.end();++assoc)
	{
		const size_t p = assoc->first;
		const size_t o = assoc->second;
		//cout << " pred #" << p << " -> obs #" << o << endl;

		win3.plot(
			_v<2,double>( observations[results.obsidx2id[o]].mean.x(), predictions[results.predidx2id[p]].mean.x() ), 
			_v<2,double>( observations[results.obsidx2id[o]].mean.y(), predictions[results.predidx2id[p]].mean.y() ), 
			"k", format("ic_%u_%u",(unsigned)p,(unsigned)o));
	}

	win3.axis_fit(true);
	win3.setPos(10,430);
	win3.resize(400,400);



	mrpt::system::pause();
}

// ------------------------------------------------------
//						MAIN
// ------------------------------------------------------
int main(int argc, char **argv)
{
	try
	{
		Test_dataAssociation( 10 );

		return 0;
	} catch (std::exception &e)
	{
		std::cout << "Exception: " << e.what() << std::endl;
		return -1;
	}
	catch (...)
	{
		printf("Untyped exception!");
		return -1;
	}
}
